@@ -15,7 +15,7 @@
  *
  */
 /*
- *  Copyright (C) 2000-2017 Thomas Habets <thomas@habets.se>
+ *  Copyright (C) 2000-2019 Thomas Habets <thomas@habets.se>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -45,6 +45,10 @@
 #include <unistd.h>
 #endif
 
+#if HAVE_GETOPT_H
+#include <getopt.h>
+#endif
+
 #if HAVE_STDINT_H
 #include <stdint.h>
 #endif
@@ -69,6 +73,10 @@
 #include <sys/param.h>
 #endif
 
+#if HAVE_SYS_RANDOM_H
+#include <sys/random.h>
+#endif
+
 #if HAVE_GRP_H
 #include <grp.h>
 #endif
@@ -124,6 +132,8 @@
 #define CLOCK_MONOTONIC CLOCK_REALTIME
 #endif
 
+#define UNUSED(x) (void)(x)
+
 /**
  * OS-specific interface finding using routing table. See findif_*.c
  * ebuf must be called with a size of at least
@@ -159,8 +169,11 @@ uint32_t dstip;
  */
 static uint8_t dstmac[ETH_ALEN];
 
+static char* payload_suffix = NULL;
+static ssize_t payload_suffix_size = -1;
+
 uint32_t srcip;                   /* autodetected, override with -S/-b/-0 */
-static uint8_t srcmac[ETH_ALEN];  /* autodetected, override with -s */
+uint8_t srcmac[ETH_ALEN];         /* autodetected, override with -s */
 
 static int16_t vlan_tag = -1; /* 802.1Q tag to add to packets. -V */
 static int16_t vlan_prio = -1; /* 802.1p prio to use with 802.1Q. -Q */
@@ -171,6 +184,7 @@ static int alsototal = 0;            /* print sent as well as received. -u */
 static int addr_must_be_same = 0;    /* -A */
 static int unsolicited = 0;          /* -U */
 static int send_reply = 0;           /* Send reply instead of request. -P */
+static int promisc = 0;              /* Use promisc mode. -p */
 
 static int finddup = 0;              /* finddup mode. -d */
 static int dupfound = 0;             /* set to 1 if dup found */
@@ -179,7 +193,6 @@ static char lastreplymac[ETH_ALEN];  /* if last different from this then dup */
 unsigned int numsent = 0;                   /* packets sent */
 unsigned int numrecvd = 0;                  /* packets received */
 static unsigned int max_replies = UINT_MAX; /* exit after -C replies */
-static unsigned int numdots = 0;            /* dots that should be printed */
 static const char* timestamp_type = NULL;   /* Incoming packet measurement ts type (-m) */
 
 static double stats_min_time = -1;
@@ -199,12 +212,47 @@ static enum { NORMAL,      /* normal output */
 static const uint8_t ethnull[ETH_ALEN] = {0, 0, 0, 0, 0, 0};
 static const uint8_t ethxmas[ETH_ALEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
 static const char* ip_broadcast = "255.255.255.255";
+static const uint32_t ip_broadcast_num = (uint32_t)-1;
 
 int verbose = 0;  /* Increase with -v */
 
 /* Doesn't really need to be volatile, but doesn't hurt. */
 static volatile sig_atomic_t time_to_die = 0;
 
+static ssize_t
+xgetrandom(void *buf, const size_t buflen, const unsigned int flags)
+{
+#ifdef HAVE_GETRANDOM
+        return getrandom(buf, buflen, flags);
+#else
+        char* p = buf;
+        for (int n = 0; n < buflen; n++) {
+                p[n] = random() & 0xff;
+        }
+        return buflen;
+#endif
+}
+
+static long int
+xrandom() {
+        const int maxtry = 10;
+        for (int c = 0; c < maxtry; c++) {
+                long int ret;
+                const ssize_t rc = xgetrandom(&ret, sizeof(ret), 0);
+                if (rc == -1) {
+                        fprintf(stderr, "arping: failed to get random bytes: %s\n", strerror(errno));
+                        continue;
+                }
+                if (sizeof(ret) != rc) {
+                        fprintf(stderr, "arping: got too few random bytes %zd, want %zd\n", rc, sizeof(ret));
+                        continue;
+                }
+                return ret;
+        }
+        fprintf(stderr, "arping: failed to get random bytes after %d tries\n", maxtry);
+        exit(1);
+}
+
 /**
  * If possible, chroot.
  *
@@ -238,6 +286,12 @@ drop_fs_root()
                 }
                 return;
         }
+        if (chdir("/")) {
+                if (verbose) {
+                        printf("arping: chdir(/): %s\n", strerror(errno));
+                }
+                return;
+        }
         if (verbose > 1) {
                 printf("arping: Successfully chrooted to %s\n", pw->pw_dir);
         }
@@ -311,13 +365,13 @@ static gid_t
 must_get_group(const char* ident)
 {
         // Special case empty string, because strtol.
-        int saved_errno;
+        int saved_errno = 0;
         if (*ident) {
                 // First try it as a name.
                 {
                         struct group* gr;
                         errno = 0;
-                        if (gr = getgrnam(ident)) {
+                        if ((gr = getgrnam(ident))) {
                                 return gr->gr_gid;
                         }
                         saved_errno = errno;
@@ -349,6 +403,9 @@ must_get_group(const char* ident)
 
 /**
  * drop all privileges.
+ *
+ * To be called as early as possible. IOW: immediately after opening
+ * raw socket.
  */
 static void
 drop_privileges(const char* drop_group)
@@ -381,6 +438,41 @@ drop_privileges(const char* drop_group)
         drop_fs_root();
         drop_uid(uid, gid);
         drop_capabilities();
+#ifdef HAVE_UNVEIL
+        if (unveil("/", "")) {
+                fprintf(stderr,
+                        "arping: failed to unveil(/, <>): %s\n",
+                        strerror(errno));
+                exit(1);
+        }
+        if (unveil(NULL, NULL)) {
+                fprintf(stderr, "arping: failed to unveil(NULL, NULL): %s\n",
+                       strerror(errno));
+                exit(1);
+        }
+        if (verbose > 1) {
+                printf("arping: Successfully unveiled\n");
+        }
+#endif
+}
+
+/**
+ * drop even more privileges, where possible.
+ *
+ * After all setup is done and main loop is about to start.
+ */
+static void
+drop_more_privileges()
+{
+#ifdef HAVE_PLEDGE
+        if (pledge("stdio tty", "")) {
+                fprintf(stderr, "arping: failed to pledge(stdio, <empty>): %s\n",
+                       strerror(errno));
+                exit(1);
+        } else if (verbose > 1) {
+                printf("arping: Successfully pledged\n");
+        }
+#endif
 }
 
 
@@ -570,7 +662,8 @@ do_libnet_init(const char *ifname, int recursive)
                         return;
                 }
                 fprintf(stderr, "arping: libnet_init(LIBNET_LINK, %s): %s\n",
-                        ifname ? ifname : "<null>", ebuf);
+                        ifname ? ifname : "<null>",
+                        *ebuf ? ebuf : "<no error message>");
                 if (getuid() && geteuid()) {
                         fprintf(stderr,
                                 "arping: you may need to run as root\n");
@@ -585,6 +678,7 @@ do_libnet_init(const char *ifname, int recursive)
 void
 sigint(int i)
 {
+        UNUSED(i);
 	time_to_die = 1;
 }
 
@@ -595,13 +689,16 @@ static void
 getclock(struct timespec *ts)
 {
 #if HAVE_CLOCK_MONOTONIC
-        if (-1 == clock_gettime(CLOCK_MONOTONIC, ts)) {
-                fprintf(stderr,
-                        "arping: clock_gettime(): %s\n",
+        static int clock_gettime_failed = 0;
+        if (!clock_gettime_failed) {
+                if (0 == clock_gettime(CLOCK_MONOTONIC, ts)) {
+                        return;
+                }
+                fprintf(stderr, "arping: clock_gettime(): %s\n",
                         strerror(errno));
-                sigint(0);
+                clock_gettime_failed = 1; // Prevent duplicate error messages.
         }
-#else
+#endif
         struct timeval tv;
         if (-1 == gettimeofday(&tv, NULL)) {
                 fprintf(stderr, "arping: gettimeofday(): %s\n",
@@ -610,7 +707,6 @@ getclock(struct timespec *ts)
         }
         ts->tv_sec = tv.tv_sec;
         ts->tv_nsec = tv.tv_usec * 1000;
-#endif
 }
 
 /**
@@ -856,7 +952,7 @@ wait_time(double deadline, uint32_t packetwait)
         if (max_wait < 0) {
                 return 0;
         }
-        if (max_wait > packetwait / 1000000) {
+        if (max_wait > packetwait / 1000000.0) {
                 return packetwait;
         }
         return max_wait * 1000000;
@@ -903,8 +999,6 @@ static char *ts2str(const struct timespec *tv, const struct timespec *tv2,
 	return buf;
 }
 
-
-
 /** Send directed IPv4 ICMP echo request.
  *
  * \param id      IP id
@@ -919,7 +1013,16 @@ pingmac_send(uint16_t id, uint16_t seq)
         // Without this padding some systems (e.g. Raspberry Pi 3
         // wireless interface) failed. dmesg said:
         //   arping: packet size is too short (42 <= 50)
-        const uint8_t padding[16] = {0};
+        const size_t padding_size = sizeof(struct timespec) + payload_suffix_size;
+        uint8_t padding[padding_size];
+        memset(padding, 0, padding_size);
+        {
+                struct timespec ts;
+                getclock(&ts);
+                memcpy(padding, &ts, sizeof(struct timespec));
+                memcpy(&padding[sizeof(struct timespec)],
+                       payload_suffix, payload_suffix_size);
+        }
 
 	int c;
 
@@ -928,7 +1031,7 @@ pingmac_send(uint16_t id, uint16_t seq)
 						   0, /* checksum */
 						   id, /* id */
 						   seq, /* seq */
-						   padding, /* payload */
+						   (uint8_t*)padding, /* payload */
 						   sizeof padding, /* payload len */
 						   libnet,
 						   icmp))) {
@@ -938,7 +1041,8 @@ pingmac_send(uint16_t id, uint16_t seq)
 	}
 
 	if (-1==(ipv4 = libnet_build_ipv4(LIBNET_IPV4_H
-					  + LIBNET_ICMPV4_ECHO_H + 0,
+                                          + LIBNET_ICMPV4_ECHO_H
+                                          + sizeof padding,
 					  0, /* ToS */
 					  id, /* id */
 					  0, /* frag */
@@ -1021,7 +1125,7 @@ pingip_send()
 					  (uint8_t*)&srcip,
 					  unsolicited ? (uint8_t*)ethxmas : (uint8_t*)ethnull,
 					  (uint8_t*)&dstip,
-					  padding,
+					  (uint8_t*)padding,
 					  sizeof padding,
 					  libnet,
 					  arp))) {
@@ -1079,14 +1183,14 @@ pingip_send()
  * \param packet  packet data
  */
 void
-pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
+pingip_recv(const char *unused, struct pcap_pkthdr *h, const char * const packet)
 {
         const unsigned char *pkt_srcmac;
         const struct libnet_802_1q_hdr *veth;
 	struct libnet_802_3_hdr *heth;
 	struct libnet_arp_hdr *harp;
         struct timespec arrival;
-	int c;
+        UNUSED(unused);
 
         if (verbose > 2) {
 		printf("arping: received response for IP ping\n");
@@ -1154,6 +1258,33 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
                 printf("arping: ... sent by acceptable host\n");
         }
 
+        // Special case: If we're not in promisc mode we could still
+        // get packets where DST mac is not us, if they're *sent* from
+        // the local host. This is an edge case but in general falls under "is promisc?".
+        //
+        // It may cause confusion because `-p` now means not just
+        // enable promisc mode (disable filter on card / in kernel),
+        // but also allow packets to any destination (disable filter
+        // in `arping`).
+        {
+                const uint8_t* p = (u_char*)harp
+                        + sizeof(struct libnet_arp_hdr)
+                        + ETH_ALEN
+                        + IP_ALEN;
+                char buf[128];
+                if (!promisc && memcmp(p, srcmac, ETH_ALEN)) {
+                        format_mac(p, buf, sizeof buf);
+                        if (verbose > 3) {
+                                printf("arping: ... but sent from %s\n", buf);
+                        }
+                        return;
+                }
+        }
+        if (verbose > 3) {
+                printf("arping: ... destination is the source we used\n");
+        }
+
+
         // Actually the IPv4 address we asked for.
         uint32_t ip;
         memcpy(&ip, (char*)harp + harp->ar_hln + LIBNET_ARP_H, 4);
@@ -1228,7 +1359,7 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
  * \param packet  packet data
  */
 void
-pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
+pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
 {
         const unsigned char *pkt_dstmac;
         const unsigned char *pkt_srcmac;
@@ -1237,7 +1368,7 @@ pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
 	struct libnet_ipv4_hdr *hip;
 	struct libnet_icmpv4_hdr *hicmp;
         struct timespec arrival;
-	int c;
+        UNUSED(unused);
 
 	if(verbose>2) {
 		printf("arping: received response for mac ping\n");
@@ -1264,6 +1395,10 @@ pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
                 return;
         }
 
+        if (verbose > 3) {
+                printf("arping: ... right dst mac\n");
+        }
+
         // Source MAC must match, if set.
         if (memcmp(dstmac, ethxmas, ETH_ALEN)) {
                 if (memcmp(pkt_srcmac, dstmac, ETH_ALEN)) {
@@ -1271,6 +1406,10 @@ pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
                 }
         }
 
+        if (verbose > 3) {
+                printf("arping: ... right src mac\n");
+        }
+
         // IPv4 Address must be me (maybe).
         if (addr_must_be_same) {
                 uint32_t tmp;
@@ -1280,11 +1419,48 @@ pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
                 }
         }
 
-        // Must be ICMP echo reply.
+        if (verbose > 3) {
+                printf("arping: ... src IP acceptable\n");
+        }
+
+        // Must be ICMP echo reply type.
         if (htons(hicmp->icmp_type) != ICMP_ECHOREPLY) {
                 return;
         }
 
+        if (verbose > 3) {
+                printf("arping: ... is echo reply type\n");
+        }
+
+        // Must be ICMP echo reply code 0.
+        if (htons(hicmp->icmp_code) != 0) {
+                return;
+        }
+
+        if (verbose > 3) {
+                printf("arping: ... is echo reply code\n");
+        }
+
+        const char* payload = (char*)hicmp + LIBNET_ICMPV4_ECHO_H;
+        const ssize_t payload_size = h->len - (payload - (char*)packet);
+        if (payload_size < 0) {
+                return;
+        }
+        if (payload_size < sizeof(struct timespec) + payload_suffix_size) {
+                return;
+        }
+        if (verbose > 3) {
+                printf("arping: ... correct payload size (%zd)\n",
+                       payload_size);
+        }
+        if (memcmp(&payload[sizeof(struct timespec)],
+                    payload_suffix, payload_suffix_size)) {
+                    return;
+        }
+        if (verbose > 3) {
+                printf("arping: ... correct payload suffix\n");
+        }
+
         update_stats(timespec2dbl(&arrival) - timespec2dbl(&lastpacketsent));
         if (beep) {
                 printf("\a");
@@ -1356,7 +1532,7 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
        struct timespec endtime;
        char done = 0;
        int fd;
-       int old_received;
+       unsigned int old_received;
 
        if (verbose > 3) {
                printf("arping: receiving packets...\n");
@@ -1505,7 +1681,6 @@ arping_main(int argc, char **argv)
 {
 	char ebuf[LIBNET_ERRBUF_SIZE + PCAP_ERRBUF_SIZE];
 	char *cp;
-	int promisc = 0;
         const char *srcip_opt = NULL;
         const char *dstip_opt = NULL;
         // `dstip_given` can be set even when there's no arg past flags on the
@@ -1521,7 +1696,7 @@ arping_main(int argc, char **argv)
         const char* drop_group = NULL;  // -g
         const char *parm; // First argument, meaning the target IP.
 	int c;
-	unsigned int maxcount = -1;
+	int maxcount = -1;
 	int dont_use_arping_lookupdev=0;
 	struct bpf_program bp;
 	pcap_t *pcap;
@@ -1530,6 +1705,7 @@ arping_main(int argc, char **argv)
         double deadline = -1;
         char bpf_filter[64];
         ebuf[0] = 0;
+        srandom(time(NULL));
 
         for (c = 1; c < argc; c++) {
                 if (!strcmp(argv[c], "--help")) {
@@ -1583,6 +1759,7 @@ arping_main(int argc, char **argv)
 			break;
 		case 'h':
 			usage(0);
+                        exit(0); // Needless but shuts up compiler warnings.
                 case 'g':
                         drop_group = optarg;
                         break;
@@ -1596,7 +1773,9 @@ arping_main(int argc, char **argv)
 					"purpose)\n");
 				exit(1);
 			}
-		case 'I': /* FALL THROUGH */
+			ifname = optarg;
+                        break;
+		case 'I':
 			ifname = optarg;
 			break;
                 case 'm':
@@ -1681,6 +1860,37 @@ arping_main(int argc, char **argv)
                 exit(1);
         }
 
+        // Generate random payload suffix for MAC pings, to be able to
+        // differentiate from unrelated ping replies.
+        if (payload_suffix_size < 0) {
+                payload_suffix_size = 4;
+                payload_suffix = malloc(payload_suffix_size);
+                if (payload_suffix) {
+                        const ssize_t rc = xgetrandom(payload_suffix, payload_suffix_size, 0);
+                        if (rc == -1) {
+                                fprintf(stderr,
+                                        "arping: failed to get %zd random bytes: %s\n",
+                                        payload_suffix_size,
+                                        strerror(errno));
+                                free(payload_suffix);
+                                payload_suffix = NULL;
+                        } else if (payload_suffix_size != rc) {
+                                fprintf(stderr,
+                                        "arping: only got %zd out of %zd bytes for random suffix\n",
+                                        rc, payload_suffix_size);
+                        }
+                } else {
+                        fprintf(stderr, "arping: failed to allocate %zd bytes for payload suffix.\n",
+                                payload_suffix_size);
+                }
+
+                if (!payload_suffix) {
+                        fprintf(stderr, "arping:  Using constant suffix.\n");
+                        payload_suffix = "arping";
+                        payload_suffix_size = strlen(payload_suffix);
+                }
+        }
+
         if (((mode == PINGIP) && opt_T)
             || ((mode == PINGMAC) && (opt_B || dstmac_opt || opt_U))) {
                 fprintf(stderr, "arping: -T can only be used to ping MAC"
@@ -1782,7 +1992,7 @@ arping_main(int argc, char **argv)
 	 * Handle dstip_given instead of ip address after parms (-B really)
 	 */
 	if (mode == NONE) {
-		if (optind + 1 == argc) {
+		if (parm) {
 			mode = is_mac_addr(parm)?PINGMAC:PINGIP;
 		} else if (dstip_given) {
 			mode = PINGIP;
@@ -1976,7 +2186,7 @@ arping_main(int argc, char **argv)
 		memcpy(srcmac, cp, ETH_ALEN);
 	}
         if (srcip_opt == NULL) {
-		if (-1 == (srcip = libnet_get_ipaddr4(libnet))) {
+		if (ip_broadcast_num == (srcip = libnet_get_ipaddr4(libnet))) {
 			fprintf(stderr,
                                 "arping: Unable to get the IPv4 address of "
                                 "interface %s:\narping: %s"
@@ -1996,6 +2206,7 @@ arping_main(int argc, char **argv)
                        format_mac(srcmac, buf, sizeof(buf)));
 	}
 
+        drop_more_privileges();
 
 	if (display == NORMAL) {
 		printf("ARPING %s\n", parm);
@@ -2010,11 +2221,9 @@ arping_main(int argc, char **argv)
                 deadline += timespec2dbl(&ts);
         }
 	if (mode == PINGIP) {
-		unsigned int c;
-                unsigned int r;
-		for (c = 0; c < maxcount && !time_to_die; c++) {
+		int c;
+		for (c = 0; (maxcount < 0 || c < maxcount) && !time_to_die; c++) {
 			pingip_send();
-                        r = numrecvd;
                         const uint32_t w = wait_time(deadline, packetwait);
                         if (w == 0) {
                                 break;
@@ -2022,11 +2231,9 @@ arping_main(int argc, char **argv)
                         ping_recv(pcap, w, (pcap_handler)pingip_recv);
 		}
 	} else { /* PINGMAC */
-		unsigned int c;
-                unsigned int r;
-		for (c = 0; c < maxcount && !time_to_die; c++) {
-			pingmac_send(rand(), c);
-                        r = numrecvd;
+		int c;
+		for (c = 0; (maxcount < 0 || c < maxcount) && !time_to_die; c++) {
+			pingmac_send(xrandom(), c);
                         const uint32_t w = wait_time(deadline, packetwait);
                         if (w == 0) {
                                 break;
